import java.util.Iterator;

int boundary = 10;
int initHuman = 50;
int initZombie = 5;
int zCounterInit = 1500;
int humanSpawns = 1;
float evolutionRate = 0.05;

color humanColor = color(250,120,120);
color zombieColor = color(160,170,90);

Crowd crowd;

void setup(){
  size(800,800);
  background(255);
  crowd = new Crowd();
  initializeCreatures(initHuman, false);
  initializeCreatures(initZombie, true);
}

void draw(){
  background(255);
  crowd.run();
}

void initializeCreatures(int num,boolean zState){
  for (int i=0;i<num;i++){
    PVector location = new PVector(random(width),random(height));
    float mspeed = random(0.5,2);
    float mforce = random(0.002,.02);
    float desSep = random(10,200);
    float vision = random(20,200);
    float sep = random(0.2,2);
    float att = random(0.2,2);
    float ali = random(0.2,2);
    float coh = random(0.2,2);
    Creature c = new Creature(zState,location, mspeed,mforce,desSep,vision, sep, att, ali, coh);
    crowd.addCreature(c, zState);
  }
}

void mouseClicked(){
  //crowd.createHuman(new PVector(mouseX,mouseY));
  //crowd.makeAllZombies();
  crowd.printStats();
}

  
class Crowd{
  
  ArrayList<Creature> creatures;
  ArrayList<Creature> humans;
  ArrayList<Creature> zombies;
  ArrayList<Creature> zombieCage;
  
  float zms, zmf, zdf, zds, zv, zs, zat, zal, zc;
  float hms, hmf, hdf, hds, hv, hs, hat, hal, hc;
  
  Crowd(){
    creatures = new ArrayList<Creature>();
    humans = new ArrayList<Creature>();
    zombies = new ArrayList<Creature>();
    zombieCage = new ArrayList<Creature>();
  }
  
  void computeAverageValues(){
    int countz = 0;
    int counth = 0;
    zms = zmf = zdf = zds = zv = zat = zal = zc = 0;
    for(Creature c :zombies){
      zms += c.maxspeed;
      zmf += c.maxforce;
      zds += c.desiredSeparation;
      zv += c.vision;
      zs += c.separate;
      zat += c.attract;
      zal += c.align;
      zc += c.cohesion;
      countz ++;
    }
    zms /= countz;
    zmf /= countz;
    zds /= countz;
    zv /= countz;
    zs /= countz;
    zat /= countz;
    zal /= countz;
    zc /= countz;
    
    hms = hmf = hdf = hds = hv = hat = hal = hc = 0;
    for(Creature h :humans){
      hms += h.maxspeed;
      hmf += h.maxforce;
      hds += h.desiredSeparation;
      hv += h.vision;
      hs += h.separate;
      hat += h.attract;
      hal += h.align;
      hc += h.cohesion;
      counth ++;
    }
    hms /= counth;
    hmf /= counth;
    hds /= counth;
    hv /= counth;
    hs /= counth;
    hat /= counth;
    hal /= counth;
    hc /= counth;
  }
  
  float rnd(float number) {
    return (float)(round((number*pow(10, 3))))/pow(10, 3);
  }

  void printStats(){
    println("h: " + humans.size() + "  z: " + zombies.size());
    println("Zombies: "+ "ms: "+rnd(zms)+",mf "+rnd(zmf)+",ds "+rnd(zds)+",v "+rnd(zv)+",s "+rnd(zs)+",at "+rnd(zat)+",al "+rnd(zal)+",zc "+rnd(zc));
    println("Humans: "+"ms: "+rnd(hms)+",mf "+rnd(hmf)+",ds "+rnd(hds)+",v "+rnd(hv)+",s "+rnd(hs)+",at "+rnd(hat)+",al "+rnd(hal)+",zc "+rnd(hc));
    println(" ");
  }
    
  void displayText(){
    noStroke();
    textSize(30);
    fill(zombieColor);
    text(zombies.size(),70,50);
    fill(humanColor);
    text(humans.size(), 20,50);
  }
    
  void run(){    
    displayText();
    //println();
    computeAverageValues();
    makeZombies();
    for(Creature h: humans){
      h.applyBehaviours(creatures);
      h.update();
    }
    Iterator<Creature> zit = zombies.iterator();
    while(zit.hasNext()){
      Creature z = zit.next();
      z.applyBehaviours(creatures);
      z.update();
      if(z.dead == true){
        if(zombies.size()>5){
          for(int i=0; i<humanSpawns; i++){
             PVector rand = new PVector(int(random(-10,10)),int(random(10,10)));
             PVector loc= PVector.add(rand,z.location);
             createHuman(loc);
          } 
          zit.remove();
        }
      }
    }
  }
  
  float addVar(float num){
    float tolerance = num*evolutionRate;
    float var = random(-tolerance, tolerance);
    num += var;
    return num;
  }

  
  void addCreature(Creature c, boolean zombieState){
    creatures.add(c);
    if (zombieState){
      zombies.add(c);
    }else{
      humans.add(c);
    }
  }
  
  void makeZombies(){
    Iterator<Creature> hit = humans.iterator();
    while(hit.hasNext()){
      Creature h = hit.next();
      for(Creature z: zombies){
        if(PVector.dist(h.location,z.location)<5){
          z.resetTimer();
          zombieCage.add(h);
          h.becomeZombie(addVar(zms), addVar(zmf),addVar(zds),addVar(zv),addVar(zs),addVar(zat),addVar(zal),addVar(zc));
          hit.remove();
          break;
        }
      }
    }
    for(Creature c: zombieCage){
      zombies.add(c);
    }
    zombieCage.clear();
  } 
  
  void createHuman(PVector pos){
    Creature spawn = new Creature(false, pos, addVar(hms),addVar(hmf),addVar(hds),addVar(hv),addVar(hs),addVar(hat),addVar(hal),addVar(hc));
    addCreature(spawn, false);        
  }
}
class Creature {
  
  //BASIC VARIABLES
  PVector location,velocity,acceleration;
  PVector diff;
  float r;
  boolean zombieState;
  boolean dead = false;
  
  //GLOBAL DNA
  float maxspeed, maxforce, desiredSeparation, vision, separate, attract, align, cohesion;
  
  //HUMAN DNA
  int humanSize = 3;
  
  
  //ZOMBIE DNA
  int zombieSize = 4;
  int zombieCounter = zCounterInit;
  
  Creature(boolean zs, PVector pos, float ms, float mf, float ds, float v, float _s, float _at, float _al, float _c){
    location = pos.get();
    acceleration = new PVector(0,0);
    velocity = new PVector(0,0);
    
    zombieState = zs; 
    
    separate = _s;
    attract = _at;
    align = _al;
    cohesion = _c;
      
    maxspeed = ms;
    maxforce = mf;
    desiredSeparation = ds;
    vision = v;
  }
  
  //HELPER FUNCTIONS
  
  void applyForce(PVector force){
    acceleration.add(force);
  }
  
  PVector seek(PVector target){
    PVector desired = PVector.sub(target,location);
    desired.normalize();
    desired.mult(maxspeed);
    PVector steer = PVector.sub(desired,velocity);
    steer.limit(maxforce);
    return steer;
  }
  
  //FLOCKING FUNCTIONS
  
  PVector align(ArrayList<Creature> creatures){
      PVector sum = new PVector(0,0);
      int count = 0;
      
      for (Creature other:creatures){
        if(zombieState == false && other.zombieState == false || zombieState ==true && other.zombieState == true){
          float d = PVector.dist(location,other.location);
          if ((d>0) && (d<vision)){
            //PVector locationDiff = PVector.sub(location, other.location);
            sum.add(other.velocity);
            count++;
          }
        }
      }
      
      if(count>0){
        sum.div(count);
        sum.normalize();
        sum.mult(maxspeed);
        PVector steer = PVector.sub(sum, velocity);
        steer.limit(maxforce);
        return steer;
      }else{
        return new PVector(0,0);
      }
  }
  
  PVector cohesion(ArrayList<Creature> creatures){
      PVector sum = new PVector(0,0);
      int count = 0;
      
      for (Creature other:creatures){
        if(zombieState == false && other.zombieState == false || zombieState ==true && other.zombieState == true){
          float d = PVector.dist(location,other.location);
          if ((d>0) && (d<vision)){
            //PVector locationDiff = PVector.sub(location, other.location);
            //float angleBetween = acos((velocity.dot(locationDiff))/(velocity.mag()*locationDiff.mag()));
            sum.add(other.location);
            count++;
          }
        }
      }
      if(count > 0){
        sum.div(count);
        return seek(sum);
      }else{
        return new PVector(0,0);
      }
  }
  
  PVector separateSimilar(ArrayList<Creature> creatures){
    PVector sum = new PVector();
    PVector steer = new PVector();
    int count = 0;
    for(Creature other: creatures){
      if(zombieState == false && other.zombieState == false || zombieState ==true && other.zombieState == true){
          float d = PVector.dist(location,other.location);
          if ((d>0) && (d<vision)){
            PVector diff = PVector.sub(other.location, location);
            diff.mult(-1);
            diff.normalize();
            diff.div(d);
            sum.add(diff);
            count++;
        }
      }
    }
    if(count>0){
      sum.div(count);
      sum.normalize();
      sum.mult(maxspeed);
      steer = PVector.sub(sum, velocity);
      steer.limit(maxforce);
    }
    return steer;
  }    
  
  PVector attractDifferent(ArrayList<Creature> creatures){
    PVector sum = new PVector();
    PVector steer = new PVector();
    int count = 0;
    for(Creature other: creatures){
      if(zombieState == false && other.zombieState == true || zombieState ==true && other.zombieState == false){
          float d = PVector.dist(location,other.location);
          if ((d>0) && (d<vision)){
            PVector diff = PVector.sub(other.location, location);
            if(!zombieState){
              diff.mult(-1);
            }
            diff.normalize();
            diff.div(d);
            sum.add(diff);
            count++;
        }
      }
    }
    if(count>0){
      sum.div(count);
      sum.normalize();
      sum.mult(maxspeed);
      steer = PVector.sub(sum, velocity);
      steer.limit(maxforce);
    }
    return steer;
  }    
  //follow/not follow flow field.
  
  //behavioral function
  
  //become zombie function
  
  void becomeZombie(float ms, float mf, float ds, float v, float _s, float _at, float _al, float _c){
    zombieState = true;
    maxspeed = ms;
    maxforce = mf;
    desiredSeparation = ds;
    vision = v;
    separate = _s;
    attract = _at;
    align = _al;
    cohesion = _c;
  }      
  
  //GLOBAL FUNCTIONS
  
  void applyBehaviours(ArrayList<Creature> creatures){
    PVector sep = separateSimilar(creatures);
    PVector att = attractDifferent(creatures);
    PVector ali = align(creatures);
    PVector coh = cohesion(creatures);
    
    sep.mult(separate);
    att.mult(attract);
    ali.mult(align);
    coh.mult(cohesion);

    applyForce(sep);
    applyForce(att);
    applyForce(ali);
    applyForce(coh);
}
    
  void update(){
    velocity.add(acceleration);
    velocity.limit(maxspeed);
    location.add(velocity);
    acceleration.mult(0);
    limitLocation();
    display();
    if(zombieState ==true){
      zombieCounter --;
      if(zombieCounter <1){
        dead = true;
      }
    }
    else{
      //HUMAN FUNCTIONS
    }
  }
  void resetTimer(){
    zombieCounter = zCounterInit;
  }
  void display(){
    pushMatrix();
    translate(location.x,location.y);
    fill(123);
    rotate(velocity.heading2D() +PI/2);
    
    noStroke();
    //println(zombieState);
    if(zombieState==true){
      r = zombieSize;
      fill(zombieColor);
    }else if(zombieState ==false){
      r = humanSize;
      fill(humanColor);
    }
    //ellipse(0,0,10,10);
    beginShape();
    vertex(0,-r*2);
    vertex(-r,r*2);
    vertex(r,r*2);
    
    endShape(CLOSE);
    popMatrix();
  }
  
  void limitLocation(){
    if(location.x>width+boundary){
      location.x = -boundary;
    }
    else if(location.x<-boundary){
      location.x = width+boundary;
    }
    if(location.y>height+boundary){
      location.y=-boundary;
    }
    else if(location.y<-boundary){
      location.y=height+boundary;
    }
  }
  
}
  
  
  
  
  

